//------------------------------------------------------------------------
//
// OpenDCC - DCC Sniffer
//
// Copyright (c) 2006, 2007, 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-04-02 V0.01 kw start (starting with the work of Alex Shephard,
//                                see www.sourceforge.net)
//            2008-04-19 V2.00 kw first beta with all functions (stable)
//            2008-05-01 V2.01 kw added drop to lf if cr-lf is received;
//                                accept both cr, lf and cr-lf
//
//------------------------------------------------------------------------
//
// purpose:   dcc sniffer
//            here: init, mainloop, globals
//
// still to do: record trigger event (reset time?)
//
//
// ====== >>>> config.h  is the central definition file for the project
//
//------------------------------------------------------------------------
//
// Open Items: Register Programming
//             Railcom
//             Polarity

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"              // general definitions the decoder, cv's
#include "hardware.h"            // port definitions for target
#include "led_control.h"         // state engine for LEDs (blinking, flashing ...)
#include "rs232.h"               // raw serial interface
#include "pc_send.h"             // envelope routines for serial interface
#include "rprintf.h"             // small printf lib from pascal stang
#include "TimerPWM.h"            // systick and PWM value
#include "DccRx.h"               // DCC receiver
#include "keyboard.h"            // only one key
#include "external_trig.h"       // watch trigger input


#include "parser.h"              // command parser
#include "tracker.h"             // decode dcc and do the analysis

// #include "main.h"

#define SIMULATION  0               // 0: real application 
                                    // 1: test receive routine
                                    // 2: test timing engine
                                    // 3: test action


void init_main(void)
  {
    DDRA  = (1 << LED_PROG)        // 0 program   - yellow
          | (1 << LED_RAILCOM)     // 1 railcom   - yellow
          | (1 << LED_ACC)         // 2 accessory - green
          | (1 << LED_LOCO)        // 3 loco      - green
          | (1 << LED_GO)          // 4 sniffing  - blue
          | (1 << LED_ON);         // 5 on        - green
    DDRB  = (1 << RC_LOWPASS_OUT)  // 1 B1 pwm for rc lowpass
          | (0 << A_THRESHOLD)     // 2 B2 either threshold or RXD1
          | (0 << A_CAPTURE);      // 3 B3 either analog capture or TXD1
    DDRC  = (0 << TRIGGER)         // input
          | (0 << SPARE);          // input on 6-pin header;
    DDRD  = (0 << MY_RXD)          // 0 input
          | (1 << MY_TXD)          // 1     // out
          | (0 << DCC_IN)          // 2     // in (Int 0)
          | (0 << MY_RTS)          // 3     // in
          | (0 << KEY)             // 4     // in (pullup)
          | (1 << MY_CTS)          // 5     // out
          | (0 << SJUMPER7)        // 6     // in (pullup)
          | (0 << SJUMPER8);       // 7     // in (pullup)       
    DDRE  = (0 << D_CAPTURE);      // 0 input

    PORTA  = (0 << LED_PROG)        // 0 program   - yellow
          | (0 << LED_RAILCOM)     // 1 railcom   - yellow
          | (0 << LED_ACC)         // 2 accessory - green
          | (0 << LED_LOCO)        // 3 loco      - green
          | (0 << LED_GO)          // 4 sniffing  - blue
          | (1 << LED_ON);         // 5 on        - green
    PORTB  = (0 << RC_LOWPASS_OUT)  // 1 B1 pwm for rc lowpass
          | (0 << A_THRESHOLD)     // 2 B2 either threshold or RXD1
          | (0 << A_CAPTURE);      // 3 B3 either analog capture or TXD1
    PORTC  = (1 << TRIGGER)         // input -> pullup
          | (1 << SPARE);          // input on 6-pin header; -> pullup
    PORTD  = (0 << MY_RXD)          // 0 input
          | (0 << MY_TXD)          // 1     // out
          | (1 << DCC_IN)          // 2     // in (Int 0)
          | (1 << MY_RTS)          // 3     // in
          | (1 << KEY)             // 4     // in (pullup)
          | (0 << MY_CTS)          // 5     // out
          | (1 << SJUMPER7)        // 6     // in (pullup)
          | (1 << SJUMPER8);       // 7     // in (pullup)       
  }


void run_echo(void)
  {
     uint8_t echo;
          if (rx_fifo_ready())
          {
            echo = rx_fifo_read();     
            led_pulse(LED_GO);
            tx_fifo_write(echo);
          }
  }


int main(void)
  {
    init_main();                // setup hardware ports

    init_led();                 // setup led states

    initDccRx();                // setup DCC receiver (uses INT0 and Timer 0)

    init_timer();               // system timer (this is timer 2)

    init_rs232(DEFAULT_BAUD);   // init_rs232(BAUD_115200);

    rprintfInit(pc_send);

    init_keyboard();            // okay, not really a _keyboard_ :-)

    init_trigger();             // watch external trigger (add-on)

    init_parser();              // command parser

    init_tracker();             // dcc analyzer

    // Some commands used for simulation
    // push_to_rx('?'); push_to_rx(0x0D); push_to_rx(0x0A);  // Help
    

    sei();

    while(1)
      {
        if (keyboard() == KEY)
          {
            toggle_EnableDisplay();
          }
        run_timer();            // watches timer isr and creates system clock
        run_led();              // watches system clock and do appropriate changes on LEDs
        
        run_parser();           // checks command input from host
        run_tracker();          // checks DCC and does the analysis

        watch_railcom();        // test railcom feedback (could be ommitted)
        
        // run_echo();          // just for testing - echo any byte from host
      }

  }


